/*
 * ESPPackU8X8.ino - "Drive" line-following car with UDP packets from UDP Joystick app
 *    (and other UDP datagram sources)
 *  target platform - ESP8266 via WeMos D1 mini NodeMCU platform
 *
 * This program receives UDP datagrams and translates them into PWM control of the
 * two motors of a slightly modified line-following "mouse". I'm using the "The D2-5
 * Intelligent Tracking Car" which is easily found for sale.
 *
 * This WORKING version now uses the U8x8 functions of the U8g2 graphics library to
 * speed writing to the onboard display. That change keeps display updates from hurting
 * car reception of datagrams.
 *
 * This code should be easily adaptable to other control applications.
 *
 * Please drop me a note if you find this code useful - jed at pobox.com Sept 2019
 *
 */

// Define PackCount to change car display to count packets per second
// Comment out define to let it display value in some received packets
//      #define PACKCOUNT 1

#include <ESP8266WiFi.h>
#include <WiFiUDP.h>

// Includes for running display
#include <SPI.h>
#include <U8g2lib.h>

#ifdef PACKCOUNT
#include <Ticker.h>  // Ticker Library
Ticker tick;
#endif
int nPackets=0;    // always count packets to keep contitional code simple
int nPackBad=0;

#define BLUELED 2  // ESP-12E onboard LED pin

char forDisp[64];  // buffer for building display items

WiFiUDP Udp;

const char* ssid     = "HomeToo";
const char* password = "password";  // NOT my password, but yours?

// const char* ssid     = "NETGEAR";
// const char* password = "";

#define LEFT_M 16     // Left motor connection
#define RIGHT_M 14    //  Right motor connection
#define UDPPORT 4949
#define MAX_PWM 1023

// where is the display attached? HERE
#define CLOCK_GPIO 5
#define DATA_GPIO 4
/* U8g2 Constructor */
// This U8X8 version does NOT use ram buffer and only writes part of screen at a time
U8X8_SSD1306_128X64_NONAME_SW_I2C u8x8(CLOCK_GPIO, DATA_GPIO, /* reset=*/ U8X8_PIN_NONE);

void setup() {
    pinMode(BLUELED, OUTPUT);
    // ready screen for display
    u8x8.begin();

    Serial.begin(115200);
    delay(10);
    u8x8.setFont(u8x8_font_8x13B_1x2_f);
    // column x, row y. A display 128x64 pixel has 16 colums and 8 rows
    u8x8.draw1x2String(0,0, "connecting");

    Serial.println();
    Serial.print("Connecting to ");
    Serial.println(ssid);

    WiFi.begin(ssid, password);

    while(WiFi.status() != WL_CONNECTED) {
      Serial.print(".");
        delay(500);
        digitalWrite(BLUELED, !digitalRead(BLUELED));  // toggle LED while connecting
    }

    Serial.println("\nWiFi connected");
    digitalWrite(BLUELED, HIGH);    // WiFi connected now, set LED off
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());

    u8x8.setFont(u8x8_font_8x13B_1x2_f);
    // column x, row y. A display 128x64 pixel has 16 colums and 8 rows.
    sprintf(forDisp, "SSID %s", ssid);
    u8x8.drawString(0,4, forDisp);
    sprintf(forDisp, "%s", WiFi.localIP().toString().c_str());
    u8x8.draw1x2String(0, 0, forDisp);
    delay(50);

    if(Udp.begin(UDPPORT) == 1){
      Serial.println("UDP connected");
      // initialize pins
      analogWrite(LEFT_M,MAX_PWM);
      analogWrite(RIGHT_M,MAX_PWM);
      } else {
      Serial.println("UDP NOT connected");
      }
      #ifdef PACKCOUNT // Needed ONLY if counting packets per second
      tick.attach(1.0, tick_print); // routine runs every tick
      #endif
}

#ifdef PACKCOUNT // Needed ONLY if counting packets per second
void tick_print()
{ // routine runs every tick
      sprintf(forDisp, "got %d packets", nPackets);
      u8x8.drawString(0,4, forDisp);
      sprintf(forDisp, "(%d wrong size)", nPackBad);
      u8x8.drawString(0,6, forDisp);
      Serial.print("Packets in interval=");
      Serial.print(nPackets);
      Serial.print(" bad=");
      Serial.println(nPackBad);
      nPackBad=nPackets=0; // restart count for next interval
} // end tick_print
  #endif

#define PKTSZ 30
int packtot, lasttot=90000;
void loop() {
  int packetSize, t_command, f_command, forward, turn, l_motor, r_motor;
  char inPac[PKTSZ+1];     // allow room for 0x00 to terminate string
  packetSize = Udp.parsePacket();
  if(packetSize > 0 ) {
    nPackets++;
    Serial.println("");
    Serial.print("Received packet of size ");
    Serial.println(packetSize);

    packetSize = Udp.read(inPac, PKTSZ);
    inPac[packetSize]=0;
    Serial.println(inPac);
    if (packetSize == 16) {
      t_command = ((inPac[0]-'0')*1000 + (inPac[1]-'0')*100 + (inPac[2]-'0')*10 + inPac[3]-'0')-1000;
      f_command = ((inPac[4]-'0')*1000 + (inPac[5]-'0')*100 + (inPac[6]-'0')*10 + inPac[7]-'0')-1000;
      if (t_command > 999) t_command = 999; // keep to 3 digits for OLED display
      if (f_command > 999) f_command = 999; // keep to 3 digits for OLED display
      Serial.print("turn=");
      Serial.print(t_command);
      Serial.print(" forward=");
      Serial.print(f_command);

      forward=(f_command-500);              // normalize speed forward to amount (forward OR back)
      if (forward < 0) forward=0;           // this robot HAS no back so set to zero
      r_motor=l_motor=forward;              // set both motors to forward "speed" 0-500
      turn=abs(t_command-500);              // compute magnitude of turn
      if (turn > forward) turn = forward;   // limit turn to motor speed == 0
      if (t_command > 500) {                // apply turn to slow correct motor
        r_motor -= turn;
      }
        else l_motor -= turn;
// now motors range from 0 to 500 but the correct range is 0 to 1023. Doubling is close.

      Serial.print("  turn=");
      Serial.print(turn);
      Serial.print("  l_motor=");
      Serial.print(l_motor);
      Serial.print("  r_motor=");
      Serial.println(r_motor);

// sinking not sourcing so reverse, change range from 0-500 to 0-1023
      analogWrite(LEFT_M,MAX_PWM-(l_motor*MAX_PWM/500.));
      analogWrite(RIGHT_M,MAX_PWM-(r_motor*MAX_PWM/500.));

#ifndef PACKCOUNT // if NOT counting packets, then display their content
// now only display packet if a LOT different from the last
      packtot=l_motor+r_motor;
  if ( abs(packtot-lasttot)> 200 ) { // if big change
      Serial.println("change");
      sprintf(inPac, "%03d,%03d", t_command,f_command);  // display UDP commands
      u8x8.setFont(u8x8_font_inr21_2x4_f );
      u8x8.drawString(0,4, inPac);
      }
    lasttot=packtot;
#endif
    } else nPackBad++;
  }
}
